/*
	File:		ColourSystem.cc

	Function:	Provides a class to help with colour space conversions to
				and from three-primary display devices.

	Author:		Andrew Willmott

*/

#include "ColourSystem.h"

/*! \class ColourSystem 

	A colour system is a way from transforming from a CIE tristimulous colour
	to a particular three-primary device. 

	The colour mapping is specified by the (x,y) chromaticities of the three
	primary colours of the system, and the chromaticity and luminance of
	the system's white point. (From this, the luminances of the primaries can
	be derived.) See Foley & van Dam, p586.

	See ColourSystem.h for various predefined colours systems and illuminants.

	Sources:	http://www.inforamp.net/~poynton/ColorFAQ.html,
				http://www.efg2.com/lab/Graphics/Colors/Chromaticity.htm

	Notes.

	CIE
		Y matches the eye's spectral efficiency curve
		thus it measures the colour's *luminance*, in candela.

		X and Z are orthogonal to luminance; varying them does not 
		affect the luminance (or perceived brightness) of the colour,
		just its hue.
		
		We can find the luminance of an RGB colour by simply
		transforming it back to XYZ space, and then taking the Y
		component.  Thus if w = toXYZ[1], then dot(w, clr) is clr's
		luminance. The Luminance() method is provided for this.

	To convert a chrominance/luminance (x, y, Y) specification into an XYZ
	colour requires:
	\verbatim
		form XYZ corresponding to x, y
			clr = [x, y, 1.0 - y - x]
		normalise to a luminance of 1.0
			clr /= y;
		multiply through by luminance
			clr *= Y;	
	\endverbatim

	Beware: converting a (say) unit luminance chroma to RGB will give
	you colours that are equally perceptually bright. This may not be
	what you want: if you try to display such a range of chromas, for
	instance, the white point will get mapped to [1, 1, 1], but bluish
	hues will be mapped so that the blue component is much bigger than
	1, because of our relative insensitivity to blue.

	In such cases, you probably want the brightest-possible device colour
	corresponding to each specified chroma, so...
	\verbatim
		form XYZ corresponding to x, y
			clr = [x, y, 1.0 - y - x]
		convert to RGB
			clr = cs.XYZToGamut(clr);
		normalise to brightest possible RGB
			clr /= MaxCmpt(clr);
	\endverbatim
*/


ColourSystem::ColourSystem(
		Chroma 		primary1,
		Chroma 		primary2,
		Chroma 		primary3,
		Chroma 		whitePoint,
		ClrReal		whiteLuminance
	)
{
	Colour	whiteXYZ, csWhite;
	
	prim0 = primary1;
	prim1 = primary2; 
	prim2 = primary3;
	white = whitePoint;
	whiteLum = whiteLuminance;
	
	toXYZ[0] = Colour(prim0[0], prim1[0], prim2[0]);
	toXYZ[1] = Colour(prim0[1], prim1[1], prim2[1]);
	toXYZ[2] = cWhite - toXYZ[0] - toXYZ[1];

	fromXYZ = inv(toXYZ);

	// now we must scale so that the white point
	// maps to the cs colour [1, 1, 1].

	whiteXYZ = Colour(white[0], white[1], 1.0 - white[0] - white[1]);
	whiteXYZ *= whiteLum / white[1];

	csWhite = fromXYZ * whiteXYZ;

	fromXYZ[0] /= csWhite[0];
	fromXYZ[1] /= csWhite[1];
	fromXYZ[2] /= csWhite[2];
	col(toXYZ, 0) *= csWhite[0];
	col(toXYZ, 1) *= csWhite[1];
	col(toXYZ, 2) *= csWhite[2];	
}

Colour ColourSystem::XYZToGamut(const Colour &cieXYZ)
/*! Transform from CIE XYZ space to the colour system's space (usually
	RGB), and saturate it to keep it within the system's gamut if
	necessary.

	The version of this at http://www.fourmilab.ch/documents/specrend
	is buggy -- its choice of clip axis can be wrong.
*/
{
	Colour      csClr = fromXYZ * cieXYZ;
	Colour      t;
	ClrReal     tMin;

    // Is the contribution of one of the primaries negative?

    if (MinCmpt(csClr) < 0.0)
    {
		// find how far we must move along the line through the white
		// point until we hit each axial plane
		t = cWhite / (cWhite - csClr);

		// find smallest +ve intercept
		tMin = 1.0;
		if (t[0] >= 0.0)                tMin = t[0];
		if (t[1] >= 0.0 && t[1] < tMin) tMin = t[1];
		if (t[2] >= 0.0 && t[2] < tMin) tMin = t[2];

        csClr = cWhite + (csClr - cWhite) * tMin;
		ClipColourZero(csClr);
        Assert(MinCmpt(csClr) >= 0.0, "bad correction");
    }

    return(csClr);
}

Colour ColourSystem::XYZToGamutClip(const Colour &cieXYZ)
{
	Colour c = XYZToGamut(cieXYZ);

	if (c[0] > 1.0) c[0] = 1.0;
	if (c[1] > 1.0) c[1] = 1.0;
	if (c[2] > 1.0) c[2] = 1.0;
	
	return(c);	
}

Colour ColourSystem::ChromaToGamut(const Chroma &cieChroma)
{
	Colour c;

	c = XYZToGamut(Colour(cieChroma, 1.0 - cieChroma[0] - cieChroma[1]));
	c /= cieChroma[1];
	
	return(c);
}

Bool ColourSystem::XYZIsOutOfGamut(const Colour &cieXYZ)
{
	return(MinCmpt(fromXYZ * cieXYZ) < 0.0);
}

Bool ColourSystem::ChromaIsOutOfGamut(const Chroma &cieChroma)
{
	return(MinCmpt(fromXYZ * Colour(cieChroma, 1.0 - cieChroma[0] - cieChroma[1]))
		 < 0.0);
}

// --- CIE standard illuminants ---------------------------------------

const Chroma csIllumA(0.448, 0.407);			// Incandescent light
const Chroma csIllumB(0.348, 0.352);			// Solar light
const Chroma csIllumC(0.3101, 0.3162);			// For NTSC television
const Chroma csIllumD55(0.332, 0.348);			// Photography
const Chroma csIllumD65(0.3127, 0.3291);		// Overcast Daylight
const Chroma csIllumD75(0.299, 0.315);
const Chroma csIllumE(0.3333333, 0.3333333);	// Uniform illuminant
											
// luminances of standard illuminants, in candela/watt

const ClrReal csMaxLum(683.0);          // defined maximum at 555 nm 
const ClrReal csWhiteLum(179.0);        // uniform white light
const ClrReal csD65Lum(203.0);
const ClrReal csALum(160.0);
const ClrReal csBLum(208.0);


// NOTE: most of this data is simply reformatted from the delphi
// source code provided by EFG's fabulous CIE page above.

// The following define the x and y coordinates of the phosphors and
// reference white points of various broadcast systems.

// [Martindale91] notes that NTSC primaries aren't remotely similar
// to those used in modern displays.
//
// CCIR Report 476-1
// "Colorimetric Standards in Colour Television"
// See http://www.igd.fhg.de/icib/tv/ccir/rep_476-1/gen.html
//
// CCIR Report 624-4
// "Characteristics of Television Systems"
// See http://www.igd.fhg.de/icib/tv/ccir/rep_624/read.html

ColourSystem csNTSC(
	Chroma(0.67, 0.33),
	Chroma(0.21, 0.71),
	Chroma(0.14, 0.08),
	Chroma(0.310, 0.316)
);

// CCIR Report 476-1
// "Colorimetric Standards in Colour Television"
// See http://www.igd.fhg.de/icib/tv/ccir/rep_476-1/gen.html
//
// CCIR Report 624-4
// "Characteristics of Television Systems"
// See http://www.igd.fhg.de/icib/tv/ccir/rep_624/read.html

ColourSystem csPAL_SECAM(
	Chroma(0.64, 0.33),
	Chroma(0.29, 0.60),
	Chroma(0.15, 0.06),
	Chroma(0.313, 0.329)
);

// CCIR Recommendation 709
// "Basic Parameter Values for the HDTV Standard for the Studio and for
// International Programme Exchange"
// See http://www.igd.fhg.de/icib/tv/ccir/rec_709/pictures/page-2.tiff

// EBU Technical Document 3271 (EBU = European Broadcasting Union)
// "Interlaced version of the 1250/50 HDTV production standard"
// http://www.igd.fhg.de/icib/tv/org/ebu/td_3271/pictures/page5.tiff

// This is used for all HDTV stuff.

ColourSystem csEBU(
	Chroma(0.640, 0.330),
	Chroma(0.300, 0.600),
	Chroma(0.150, 0.060),
	Chroma(0.3127, 0.3290)
);

// SMPTE 240M (SMPTE = The Society of Motion Picture and Television Engineers)
// "Signal Parameters -- 1125-line High-Definition Production System"
// http://www.igd.fhg.de/icib/tv/org/smpte/st_240M-1992/read.html

ColourSystem csSMPTE(
		Chroma(0.630, 0.340),
		Chroma(0.310, 0.595),
		Chroma(0.155, 0.070),
		csIllumD65
);


ColourSystem csShortPersistencePhosphors(  // [Foley96, p. 583]
	Chroma(0.61, 0.35),
	Chroma(0.29, 0.59),
	Chroma(0.15, 0.063),
	csIllumC
);

ColourSystem csLongPersistencePhosphors(   // [Foley96, p. 583]
	Chroma(0.62, 0.33),
	Chroma(0.21, 0.685),
	Chroma(0.15, 0.063),
	csIllumC
);

ColourSystem csDellPhosphors(  	// 12 Jan 99 E-mail from Dell
	Chroma(0.625, 0.340),     	// All Dell monitors except
	Chroma(0.275, 0.605),     	// Nokia 91862
	Chroma(0.150, 0.065),
	csIllumD65
);


ColourSystem &csDisplay = csEBU;

